
#include "osd.h"

/* Display timing data */
int frame_skip              = 1;
int frame_count             = 0;
volatile int tick_count     = 0;
volatile int old_tick_count = 0;
int skip                    = 0;

/* Blitter data */
BITMAP *sms_bmp = NULL;
int blitter_index;
void (*blitter_proc)(BITMAP *src, BITMAP *dst) = NULL;

/* Configurable options */
int joy_driver              = JOY_TYPE_NONE;
int video_driver            = GFX_AUTODETECT;
int video_width             = 320;
int video_height            = 200;
int depth                   = 8;
int use_blur                = 0;
int use_scale               = 0;
int no_vga                  = 0;
int no_mmx                  = 0;
int use_expand              = 0;
int use_scanlines           = 0;
int use_vsync               = 0;
int use_throttle            = 0;
int use_sound               = 0;
int sound_card              = -1;
int sample_rate             = 22050;
int country                 = TYPE_OVERSEAS;
int fm_enable               = 0;

/* Game and SRAM file names */
char game_name[0x100];
char sram_name[0x100];

/* Joystick driver names and numbers */
struct {
    char *name;
    int id;
} joy_driver_table[] = {
    { "auto",             JOY_TYPE_AUTODETECT },
    { "none",             JOY_TYPE_NONE },
    { "standard",         JOY_TYPE_STANDARD },
    { "2pads",            JOY_TYPE_2PADS },
    { "4button",          JOY_TYPE_4BUTTON },
    { "6button",          JOY_TYPE_6BUTTON },
    { "8button",          JOY_TYPE_8BUTTON },
    { "fspro",            JOY_TYPE_FSPRO },
    { "wingex",           JOY_TYPE_WINGEX },
    { "sidewinder",       JOY_TYPE_SIDEWINDER },
    { "gamepdpro",        JOY_TYPE_GAMEPAD_PRO },
    { "grip",             JOY_TYPE_GRIP },
    { "grip4",            JOY_TYPE_GRIP4 },
    { "sneslpt1",         JOY_TYPE_SNESPAD_LPT1 },
    { "sneslpt2",         JOY_TYPE_SNESPAD_LPT2 },
    { "sneslpt3",         JOY_TYPE_SNESPAD_LPT3 },
    { "psxlpt1",          JOY_TYPE_PSXPAD_LPT1 },
    { "psxlpt2",          JOY_TYPE_PSXPAD_LPT2 },
    { "psxlpt3",          JOY_TYPE_PSXPAD_LPT3 },
    { "n64lpt1",          JOY_TYPE_N64PAD_LPT1 },
    { "n64lpt2",          JOY_TYPE_N64PAD_LPT2 },
    { "n64lpt3",          JOY_TYPE_N64PAD_LPT3 },
    { "db9lpt1",          JOY_TYPE_DB9_LPT1 },
    { "db9lpt2",          JOY_TYPE_DB9_LPT2 },
    { "db9lpt3",          JOY_TYPE_DB9_LPT3 },
    { "tglpt1",           JOY_TYPE_TURBOGRAFX_LPT1 },
    { "tglpt2",           JOY_TYPE_TURBOGRAFX_LPT2 },
    { "tglpt3",           JOY_TYPE_TURBOGRAFX_LPT3 },
    { "wingwar",          JOY_TYPE_WINGWARRIOR },
    { "segaisa",          JOY_TYPE_IFSEGA_ISA},
    { "segapci",          JOY_TYPE_IFSEGA_PCI},
    { "segapci2",         JOY_TYPE_IFSEGA_PCI_FAST},
    { 0, 0 }
};

int main (int argc, char **argv)
{
    /* Show version information if no arguments are specified */
    if(argc < 2)
    {
        printf("\nSMS Plus\n");
        printf("(C) Charles Mac Donald in 1998, 1999, 2000\n");
        printf("Version %s, build date: %s, %s\n", VERSION, __DATE__, __TIME__);
        printf("Usage: sp <filename.ext> [-options]\n");
        printf("Type 'sp -help' for a summary of the available options\n");
        exit(1);
    }

    /* Show option list */
    if(stricmp(argv[1], "-help") == 0)
    {
        printf("Options:\n");
        printf(" -res <x> <y> \t set the display resolution.\n");
        printf(" -scale       \t scale display to full resolution. (slow)\n");
        printf(" -expand      \t force 512x384 or 400x300 zoomed display.\n");
        printf(" -nommx       \t disable use of MMX instructions.\n");
        printf(" -novga       \t disable use of VGA vertical scaling with '-expand'.\n");
        printf(" -depth <n>   \t specify color depth. (8, 16)\n");
        printf(" -blur        \t blur display. (16-bit color only)\n");
        printf(" -scanlines   \t use scanlines effect.\n");
        printf(" -vsync       \t wait for vertical sync before blitting.\n");
        printf(" -throttle    \t limit updates to 60 frames per second.\n");
        printf(" -sound       \t enable sound. (auto-throttle)\n");
        printf(" -sndrate <n> \t specify sound rate. (8000, 11025, 22050, 44100)\n");
        printf(" -sndcard <n> \t specify sound card. (0-7)\n");
        printf(" -joy <s>     \t specify joystick type.\n");
        printf(" -jp          \t use Japanese console type.\n");
        printf(" -fm          \t required to enable YM2413 sound.\n");
        exit(1);
    }

    /* Do configuration */
    do_config();

    /* Make copy of game filename */
    strcpy(game_name, argv[1]);

    /* Attempt to load game off commandline */
    if(load_rom(game_name) == 0)
    {
        printf("Error loading `%s'.\n", game_name);
        exit(1);
    }

    /* Figure out game image type */
    if(stricmp(strrchr(game_name, '.'), ".gg") == 0)
        cart.type = TYPE_GG;
    else
        cart.type = TYPE_SMS;

    /* Make SRAM filename */
    strcpy(sram_name, game_name);
    strcpy(strrchr(sram_name, '.'), ".sav\0");

    /* Set up video, audio, input, etc. */
    init_machine();

    /* Initialize the virtual console emulation */
    system_init(use_sound ? sample_rate : 0);
    sms.country = country;
    sms.fm_enable = fm_enable;

    /* Load SRAM from disk if possible */
    load_sram(sram_name);

    /* Pick blitter function to use */
    {
        int d = (depth == 16) ? 1 : 0;
        int v = (no_vga) ? 1 : 0;
        int m = (no_mmx) ? 0 : 1;
        int e = (use_expand) ? 1 : 0;
        int z = (use_scale) ? 1 : 0;
        int s = (use_scanlines) ? 1 : 0;
        int g = (cart.type == TYPE_GG) ? 1 : 0;

        /* Disable zooming if expansion is on */
        if(e && z) z = 0;

        /* Ignore VGA and MMX disable switches if not using expansion */
        if(!e) { v = m = 0; };

        /* If expansion and scanlines are on, ignore VGA disable */
        if(e && s) v = 0;

        blitter_index = (d << 6 | v << 5 | m << 4 | e << 3 | z << 2 | s << 1 | g) & 0x7F;
        blitter_proc = blit_table[blitter_index];

        if(blitter_proc == NULL)
        {
            set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
            printf("Illegal combination of parameters - not all options can be combined.\n");
            exit(1);
        }
    }

    /* Main emulation loop */
    for(;;)
    {
        /* Bump frame count */
        frame_count += 1;

        /* Are we skipping this frame? */
        skip = (frame_count % frame_skip == 0) ? 0 : 1;

        /* Get current input */
        update_input();

        /* Set frame skip level */
        if(check_key(KEY_F1)) frame_skip = 1;
        if(check_key(KEY_F2)) frame_skip = 2;
        if(check_key(KEY_F3)) frame_skip = 3;
        if(check_key(KEY_F4)) frame_skip = 4;

        if(check_key(KEY_F5))
        {
            FILE *fd = fopen("state.sta","wb");
            if(!fd) exit(1);
            save_state(fd);
            fclose(fd);
        }

        if(check_key(KEY_F6))
        {
            FILE *fd = fopen("state.sta","rb");
            if(!fd) exit(1);
            load_state(fd);
            fclose(fd);
        }

        /* Quit if needed */
        if(key[KEY_ESC] || key[KEY_END]) break;

        /* Run the system emulation for a frame */
        sms_frame(skip);

        /* Update the display */
        if(skip == 0)
        {
            /* Play the sound stream if enabled */
            if(use_sound) osd_play_streamed_sample_16(0, snd.buffer, snd.bufsize * 2, sample_rate, 60, 0);

            /* Wait for VSync */
            if(use_vsync) vsync();

            /* Update the palette */
            if(depth == 8) update_palette();

            /* Use the blur effect */
            if(use_blur)
            {
                if(cart.type == TYPE_GG)
                    blur(&sms_bmp->line[24][96], 160, 144, 192);
                else
                    blur(&sms_bmp->line[0][0], 256, 192, 0);
            }

            blitter_proc(sms_bmp, screen);
        }

        /* Speed throttling */
        if(use_throttle)
        {
            while(tick_count == old_tick_count);
            old_tick_count = tick_count;
        }
    }

    /* Save SRAM contents to disk */
    save_sram(sram_name);

    system_shutdown();

    /* Free system resources */
    trash_machine();
    return (0);
}


/****************************************************************************/


/* Timer handler, runs at 60Hz */
void tick_handler(void)
{
    tick_count += 1;
}
END_OF_FUNCTION(tick_handler);


/* Update keyboard and joystick input */
void update_input(void)
{
    /* Clear all button states */
    memset(&input, 0, sizeof(t_input));

    /* System buttons */
    if(key[KEY_ENTER] || key[KEY_SPACE]) input.system |= (cart.type == TYPE_GG) ? INPUT_START : INPUT_PAUSE;
    if(key[KEY_DEL])                     input.system |= INPUT_HARD_RESET;
    if(key[KEY_TAB])                     input.system |= (cart.type == TYPE_GG) ? INPUT_HARD_RESET : INPUT_SOFT_RESET;

    /* Player 1 buttons */
    if(key[KEY_UP])         input.pad[0] |= INPUT_UP;
    else
    if(key[KEY_DOWN])       input.pad[0] |= INPUT_DOWN;
    if(key[KEY_LEFT])       input.pad[0] |= INPUT_LEFT;
    else
    if(key[KEY_RIGHT])      input.pad[0] |= INPUT_RIGHT;
    if(key[KEY_LCONTROL])   input.pad[0] |= INPUT_BUTTON2;
    if(key[KEY_ALT])        input.pad[0] |= INPUT_BUTTON1;

    /* Is the joystick being used ? */
    if(joy_driver != JOY_TYPE_NONE)
    {
        poll_joystick();

        /* Check player 1 joystick */
        if(joy[0].stick[0].axis[1].d1) input.pad[0] |= INPUT_UP;
        else
        if(joy[0].stick[0].axis[1].d2) input.pad[0] |= INPUT_DOWN;

        if(joy[0].stick[0].axis[0].d1) input.pad[0] |= INPUT_LEFT;
        else
        if(joy[0].stick[0].axis[0].d2) input.pad[0] |= INPUT_RIGHT;

        if(joy[0].button[0].b)  input.pad[0] |= INPUT_BUTTON2;
        if(joy[0].button[1].b)  input.pad[0] |= INPUT_BUTTON1;
        if(joy[0].button[2].b)  input.system |= (cart.type == TYPE_GG) ? INPUT_START : INPUT_PAUSE;

        /* More than one joystick supported ? */
        if(num_joysticks > 2)
        {
            /* Check player 2 joystick */
            if(joy[1].stick[0].axis[1].d1) input.pad[1] |= INPUT_UP;
            else
            if(joy[1].stick[0].axis[1].d2) input.pad[1] |= INPUT_DOWN;

            if(joy[1].stick[0].axis[0].d1) input.pad[1] |= INPUT_LEFT;
            else
            if(joy[1].stick[0].axis[0].d1) input.pad[1] |= INPUT_RIGHT;

            if(joy[1].button[0].b)  input.pad[1] |= INPUT_BUTTON2;
            if(joy[1].button[1].b)  input.pad[1] |= INPUT_BUTTON1;
        }
    }
}


/* Initialize machine dependant stuff */
void init_machine(void)
{
    /* Set up sound hardware */
    if(use_sound)
    {
        /* Initialize sound and update sample rate */
        int ret = msdos_init_sound(&sample_rate, sound_card);

        /* Error - disable sound */
        if(ret != 0) use_sound = 0;
    }

    /* Set up the Allegro library */
    allegro_init();

    /* Disable MMX usage if requested */
    if(no_mmx == 1) cpu_mmx = 0;

    /* Install input devices */
    install_keyboard();
    install_joystick(joy_driver);

    /* Install tick handler */
    install_timer();
    LOCK_FUNCTION(tick_handler);
    LOCK_VARIABLE(tick_count);
    install_int_ex(tick_handler, BPS_TO_TIMER(60));

    /* Change video settings if screen expansion is enabled */
    if(use_expand)
    {
        video_driver = GFX_AUTODETECT;
        video_width  = (cart.type == TYPE_GG) ? 400 : 512;
        video_height = (cart.type == TYPE_GG) ? 300 : 384;
    }

    /* Double requested height if scanlines are being used
       and screen expansion is disabled */
    if((use_scanlines == 1) && (use_expand == 0))
    {
        video_height *= 2;            
    }

    /* Attempt to set graphics mode */
    set_color_depth(depth);
    if(set_gfx_mode(video_driver, video_width, video_height, 0, 0) != 0)
    {
        printf("Error: %s\n", allegro_error);
        exit(1);
    }

    /* Clear palette and display */
    set_palette(black_palette);
    clear(screen);

    /* Use vertical stretching if screen expansion is enabled,
       VGA expansion is disabled, and scanlines are disabled */
    if((use_expand == 1) && (no_vga == 0) && (use_scanlines == 0))
    {
        outp(0x3D4, 0x09);
        outp(0x3D5, (inp(0x3D5) & 0xE0) | 0x03);
    }

    /* Force overscan color to be black */
    if(depth == 8)
    {
        inp(0x3DA);
        outp(0x3C0, 0x31);
        outp(0x3C0, 0xFF);
        clear_to_color(screen, 0xFF);
    }

    /* Allocate memory bitmap for blitting */
    sms_bmp = create_bitmap_ex(depth, 256, 192);
    clear(sms_bmp);

    /* Set up bitmap structure */
    memset(&bitmap, 0, sizeof(t_bitmap));
    bitmap.width  = sms_bmp->w;
    bitmap.height = sms_bmp->h;
    bitmap.depth  = depth;
    bitmap.pitch  = (bitmap.width * (bitmap.depth >> 3));
    bitmap.data   = (unsigned char *)&sms_bmp->line[0][0];
}


/* Free system resources */
void trash_machine(void)
{
    if(use_sound) msdos_shutdown_sound();
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
}


/* Update VGA palette hardware */
void update_palette(void)
{
    RGB color;
    int index;

    if(bitmap.pal.update == 0) return;

    for(index = 0; index < 32; index += 1)
    {
        if(bitmap.pal.dirty[index])
        {
            color.r = (bitmap.pal.color[index][0] >> 2);
            color.g = (bitmap.pal.color[index][1] >> 2);
            color.b = (bitmap.pal.color[index][2] >> 2);

            set_color(0x00 | index, &color);
            set_color(0x20 | index, &color);
            set_color(0x40 | index, &color);
        }
    }

    bitmap.pal.update = 0;
    memset(bitmap.pal.dirty, 0, 32);
}


/* Load a game file */
int load_rom(char *filename)
{
    int size;

    if(stricmp(strrchr(filename, '.'), ".zip") == 0)
    {
        char name[0x100];

        cart.rom = loadzip(filename, name, &size);
        if(!cart.rom) return (0);
        printf("Loaded `%s', (%d bytes) from archive `%s'\n", name, size, filename);
        strcpy(game_name, name);

        /* Take care of image header, if present */
        if((size / 512) & 1)
        {
            size -= 512;
            memmove(cart.rom, cart.rom + 512, size);
        }

        cart.pages = (size / 0x4000);
    }
    else
    {
        FILE *fd = NULL;

        fd = fopen(filename, "rb");
        if(!fd) return (0);

        /* Seek to end of file, and get size */
        fseek(fd, 0, SEEK_END);
        size = ftell(fd);
        fseek(fd, 0, SEEK_SET);

        /* Don't load games smaller than 32k */
        if(size < 0x8000) return (0);

        /* Take care of image header, if present */
        if((size / 512) & 1)
        {
            size -= 512;
            fseek(fd, 512, SEEK_SET);
        }

        cart.pages = (size / 0x4000);
        cart.rom = malloc(size);
        if(!cart.rom) return (0);
        fread(cart.rom, size, 1, fd);

        fclose(fd);
    }

    return (1);
}

void system_load_sram(void)
{
    load_sram(sram_name);
}


/* Load SRAM data */
void load_sram(char *filename)
{
    FILE *fd = fopen(filename, "rb");
    if(fd)
    {
        sms.save = 1;
        fread(sms.sram, 0x8000, 1, fd);
        fclose(fd);
    }
}


/* Save SRAM data */
void save_sram(char *filename)
{
    if(sms.save)
    {
        FILE *fd = fopen(sram_name, "wb");
        if(fd)
        {
            fwrite(sms.sram, 0x8000, 1, fd);
            fclose(fd);
        }
    }
}


/* Parse configuration file */
int parse_file(const char *filename, int *argc, char **argv)
{
    char token[0x100];
    FILE *handle = NULL;

    *argc = 0;

    handle = fopen(filename, "r");
    if(!handle) return (0);

    fscanf(handle, "%s", &token[0]);
    while(!(feof(handle)))
    {
        int size = strlen(token) + 1;
        argv[*argc] = malloc(size);
        if(!argv[*argc]) return (0);
        strcpy(argv[*argc], token);
        *argc += 1;
        fscanf(handle, "%s", &token[0]);
    }

    if(handle) fclose(handle);
    return (1);
}


void do_config(void)
{
    /* The commandline */
    extern int __crt0_argc;
    extern char **__crt0_argv;

    /* Our token list */
    int i, argc;
    char *argv[0x100];

    for(i = 0; i < 0x100; i += 1) argv[i] = NULL;

    /* Check configuration file */
    parse_file("sp.cfg", &argc, argv);

    /* Check extracted tokens */
    parse_args(argc, argv);

    /* Free token list */
    for(i = 0; i < argc; i += 1)
    {
        if(argv[argc]) free (argv[argc]);
    }

    /* Check command line */
    parse_args(__crt0_argc, __crt0_argv);
}

/* Parse argument list */
void parse_args(int argc, char **argv)
{
    int i;

    for(i = 0; i < argc; i += 1)
    {
        if(stricmp(argv[i], "-novga") == 0)
        {
            no_vga = 1;
        }

        if(stricmp(argv[i], "-nommx") == 0)
        {
            no_mmx = 1;
        }

        if(stricmp(argv[i], "-scanlines") == 0)
        {
            use_scanlines = 1;
        }

        if(stricmp(argv[i], "-scale") == 0)
        {
            use_scale = 1;
        }

        if(stricmp(argv[i], "-expand") == 0)
        {
            use_expand = 1;
        }

        if(stricmp(argv[i], "-vsync") == 0)
        {
            use_vsync = 1;
        }

        if(stricmp(argv[i], "-sound") == 0)
        {
            use_sound = 1;
        }

        if(stricmp(argv[i], "-sndcard") == 0)
        {
            sound_card = atoi(argv[i+1]);
        }

        if(stricmp(argv[i], "-sndrate") == 0)
        {
            sample_rate = atoi(argv[i+1]);
        }

        if(stricmp(argv[i], "-res") == 0)
        {
            video_width  = atoi(argv[i+1]);
            video_height = atoi(argv[i+2]);
        }

        if(stricmp(argv[i], "-depth") == 0)
        {
            depth = atoi(argv[i+1]);
        }

        if(stricmp(argv[i], "-blur") == 0)
        {
            use_blur = 1;
        }

        if(stricmp(argv[i], "-jp") == 0)
        {
            country = TYPE_DOMESTIC;
        }

        if(stricmp(argv[i], "-fm") == 0)
        {
            fm_enable = 1;
        }

        if(stricmp(argv[i], "-throttle") == 0)
        {
            use_throttle = 1;
        }

        if(stricmp(argv[i], "-joy") == 0)
        {
            int j;
            for(j = 0; joy_driver_table[j].name != 0; j += 1)
            {
                if(stricmp(argv[i + 1], joy_driver_table[j].name) == 0)
                {
                    joy_driver = joy_driver_table[j].id;                                        
                }
            }
        }
    }
}


/* Debounce keyboard buttons */
int check_key(int code)
{
    static int lastpressed = 0;

    if((!key[code]) && (lastpressed == code))
        lastpressed = 0;

    if((key[code]) && (lastpressed != code))
    {
        lastpressed = code;
        return (1);
    }

    return (0);
}


void blit_sms(BITMAP *src, BITMAP *dst) {
    blit(sms_bmp, screen, 0, 0, (SCREEN_W-256)/2, (SCREEN_H-192)/2, 256, 192);
}

void blit_sms_scanlines(BITMAP *src, BITMAP *dst) {
    int i;
    for(i=0;i<192;i+=1)
    blit(sms_bmp, screen, 0, i, (SCREEN_W-256)/2, (((SCREEN_H/2)-192))+(i << 1), 256, 1);
}

void blit_gg(BITMAP *src, BITMAP *dst) {
    blit(sms_bmp, screen, 48, 24, (SCREEN_W-160)/2, (SCREEN_H-144)/2, 160, 144);
}

void blit_gg_scanlines(BITMAP *src, BITMAP *dst) {
    int i;
    for(i=0;i<144;i+=1)
    blit(sms_bmp, screen, 48, 24+i, (SCREEN_W-160)/2, (((SCREEN_H/2)-144))+(i << 1), 160, 1);
}

void blit_sms_scale(BITMAP *src, BITMAP *dst) {
    stretch_blit(sms_bmp, screen, 0, 0, 256, 192, 0, 0, SCREEN_W, SCREEN_H);
}

void blit_sms_scale_scanlines(BITMAP *src, BITMAP *dst) {
    int i;
    for(i=0;i<192;i+=1)
    stretch_blit(sms_bmp, screen, 0, i, 256, 1, 0, (((SCREEN_H/2)-192))+(i << 1), SCREEN_W, 1);
}

void blit_gg_scale(BITMAP *src, BITMAP *dst) {
    stretch_blit(sms_bmp, screen, 48, 24, 160, 144, 0, 0, SCREEN_W, SCREEN_H);
}

void blit_gg_scale_scanlines(BITMAP *src, BITMAP *dst) {
    int i;
    for(i=0;i<144;i+=1)
    stretch_blit(sms_bmp, screen, 48, 24+i, 160, 1, 0, (((SCREEN_H/2)-144))+(i << 1), SCREEN_W, 1);
}


#include "blittbl.c"



