/*
    Copyright (C) 1998, 1999, 2000  Charles Mac Donald

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "shared.h"


t_bitmap bitmap;
t_cart cart;                
t_snd snd;
t_input input;
FM_OPL *ym3812;


void system_init(int rate)
{
    /* Initialize the VDP emulation */
    vdp_init();

    /* Initialize the SMS emulation */
    sms_init();

    /* Initialize the look-up tables and related data */
    render_init();

    /* Enable sound emulation if the sample rate was specified */
    if(rate)
    {
        /* Clear sound context */
        memset(&snd, 0, sizeof(t_snd));

        /* Calculate buffer size in samples */
        snd.bufsize = (rate / FRAMES_PER_SECOND);

        /* Sound output */
        snd.buffer = (signed short int *)malloc(snd.bufsize * 2);

        /* YM3812/YM2413 sound stream */
        snd.fm_buffer = (signed short int *)malloc(snd.bufsize * 2);

        /* SN76489 sound stream */
        snd.psg_buffer = (signed short int *)malloc(snd.bufsize * 2);

        /* Ensure we could allocate all the sound buffers */
        if(snd.buffer && snd.fm_buffer && snd.psg_buffer)
        {
            /* Set up SN76489 emulation */
            SN76496_init(0, MASTER_CLOCK, 255, rate);

            /* Set up YM3812 emulation */
            ym3812 = OPLCreate(OPL_TYPE_YM3812, MASTER_CLOCK, rate);

            /* Set up YM2413 emulation */
            ym2413_init(1);

            /* Inform other functions that we can use sound */
            snd.enabled = 1;
        }
        else
        {
            /* Free whatever buffers we were able to get up until
               running out of memory. */
            if(snd.fm_buffer) free(snd.fm_buffer);
            if(snd.psg_buffer) free(snd.psg_buffer);
            if(snd.buffer) free(snd.buffer);
        }
    }

    /* Don't save SRAM by default */
    sms.save = 0;

    /* Clear emulated button state */
    memset(&input, 0, sizeof(t_input));
}


void system_shutdown(void)
{
    if(snd.enabled) OPLDestroy(ym3812);
}


void system_reset(void)
{
    vdp_reset();
    sms_reset();
    render_reset();
    system_load_sram();
    if(snd.enabled)
    {
        OPLResetChip(ym3812);
        ym2413_reset(0);
    }
}


void save_state(void *fd)
{
    /* Save VDP context */
    fwrite(&vdp, sizeof(t_vdp), 1, fd);

    /* Save SMS context */
    fwrite(&sms, sizeof(t_sms), 1, fd);

    /* Save Z80 context */
    fwrite(Z80_Context, sizeof(Z80_Regs), 1, fd);
}


void load_state(void *fd)
{
    int i;

    /* Reset the virtual console */
    system_reset();

    /* Load VDP context */
    fread(&vdp, sizeof(t_vdp), 1, fd);

    /* Load SMS context */
    fread(&sms, sizeof(t_sms), 1, fd);

    /* Load Z80 context */
    fread(Z80_Context, sizeof(Z80_Regs), 1, fd);

    /* Set up IRQ callback function */
    z80_set_irq_callback(&irq_callback);

    /* Restore program banks */
    sms.read_map[0] = cart.rom + 0x0000;
    sms.read_map[1] = cart.rom + 0x2000;
    sms.read_map[2] = cart.rom + 0x4000;
    sms.read_map[3] = cart.rom + 0x6000;
    sms.read_map[4] = cart.rom + 0x0000;
    sms.read_map[5] = cart.rom + 0x2000;
    sms.read_map[6] = sms.ram;            
    sms.read_map[7] = sms.ram;

    sms.write_map[0] = sms.dummy;         
    sms.write_map[1] = sms.dummy;
    sms.write_map[2] = sms.dummy;         
    sms.write_map[3] = sms.dummy;
    sms.write_map[4] = sms.dummy;         
    sms.write_map[5] = sms.dummy;
    sms.write_map[6] = sms.ram;           
    sms.write_map[7] = sms.ram;           

    for(i = 0; i < 4; i += 1)
    {
        sms_mapper_w(i, sms.fcr[i]);
    }

    /* Restore pattern data */
    is_vram_dirty = 1;
    memset(vram_dirty, 1, 0x200);

    /* Restore palette data */
    for(i = 0; i < 32; i += 1)
    {
        palette_sync(i);
    }
}


