
#include "shared.h"

/* SMS system context */
t_sms sms;


void sms_frame(int skip_render)
{
    /* Do a 'hard' reset for the Game Gear */
    if((cart.type == TYPE_GG) && (input_mask & KEY_PAD_RESET))
    {
        system_reset();
    }

    /* NTSC frame has 262 lines */
    for(vdp.line = 0; vdp.line < 262; vdp.line += 1)
    {
        /* Debounce pause key */
        if(input_mask & KEY_PAD_PAUSE)
        {
            if(!sms.paused)
            {
                sms.paused = 1;

                z80_set_nmi_line(ASSERT_LINE);
                z80_set_nmi_line(CLEAR_LINE);
            }
        }
        else
        {
            sms.paused = 0;
        }

        /* Handle VDP line events */
        vdp_run();

        /* Draw the current frame */
        if(skip_render == 0) render_line(vdp.line);

        /* Run the Z80 for a line */
        z80_execute(227);
    }

    /* Update the emulated sound stream */
    if(snd.buffer)
    {
        SN76496Update(0, snd.buffer, snd.bufsize);
    }
}


void sms_init(void)
{
#if PSX
    z80_set_context((void *)0x1F800000);
#endif

    sms_reset();
}


void sms_reset(void)
{
    memset(&sms, 0, sizeof(t_sms));

    z80_reset(0);
    z80_set_irq_callback(irq_callback);

    sms.read_map[0] = cart.rom + 0x0000;
    sms.read_map[1] = cart.rom + 0x2000;
    sms.read_map[2] = cart.rom + 0x4000;
    sms.read_map[3] = cart.rom + 0x6000;
    sms.read_map[4] = cart.rom + 0x0000;
    sms.read_map[5] = cart.rom + 0x2000;
    sms.read_map[6] = sms.ram;            
    sms.read_map[7] = sms.ram;

    sms.write_map[0] = sms.dummy;         
    sms.write_map[1] = sms.dummy;
    sms.write_map[2] = sms.dummy;         
    sms.write_map[3] = sms.dummy;
    sms.write_map[4] = sms.dummy;         
    sms.write_map[5] = sms.dummy;
    sms.write_map[6] = sms.ram;           
    sms.write_map[7] = sms.ram;           
}


/* Read an opcode */
int cpu_readop(int address)
{
    return (sms.read_map[(address >> 13)][(address & 0x1FFF)]);
}


/* Read an operand */
int cpu_readop_arg(int address)
{
    return (sms.read_map[(address >> 13)][(address & 0x1FFF)]);
}


/* Read from memory */
int cpu_readmem16(int address)
{
    return (sms.read_map[(address >> 13)][(address & 0x1FFF)]);
}


/* Write to memory */
void cpu_writemem16(int address, int data)
{
    sms.write_map[(address >> 13)][(address & 0x1FFF)] = data;
    if(address >= 0xFFFC) sms_mapper_w(address & 3, data);
}


/* Write to an I/O port */
void cpu_writeport(int port, int data)
{
    switch(port & 0xFF)
    {
        case 0x01: /* GG SIO */
        case 0x02:
        case 0x03:
        case 0x04:
        case 0x05:
            break;

        case 0x06: /* GG STEREO */
            break;

        case 0x7E: /* SN76489 PSG */
        case 0x7F:
            if(snd.buffer) SN76496Write(0, data);
            break;

        case 0xBE: /* VDP DATA */
            vdp_data_w(data);
            break;

        case 0xBD: /* VDP CTRL */ 
        case 0xBF:
            vdp_ctrl_w(data);
            break;

        case 0xF0: /* YM2413 FM */
        case 0xF1:
            break;

        case 0xF2: /* YM2413 DETECT */
            break;

        case 0x3F: /* TERRITORY CTRL. */
            break;

        default:
            break;
    }
}


/* Read from an I/O port */
int cpu_readport(int port)
{
    byte temp = 0;

    switch(port & 0xFF)
    {
        case 0x01: /* GG SIO */
        case 0x02:
        case 0x03:
        case 0x04:
        case 0x05:
            return (0x00);

        case 0x7E: /* V COUNTER */
            return (vdp_vcounter_r());
            break;

        case 0x7F: /* H COUNTER */
            return (vdp_hcounter_r());
            break;

        case 0x00: /* INPUT #2 */
            temp = 0xFF;
            if(input_mask & KEY_PAD_START) temp &= ~0x80;
            return (temp);

        case 0xC0: /* INPUT #0 */  
        case 0xDC:
            temp = 0xFF;
            if(input_mask & KEY_PAD1_UP)      temp &= ~0x01;
            if(input_mask & KEY_PAD1_DOWN)    temp &= ~0x02;
            if(input_mask & KEY_PAD1_LEFT)    temp &= ~0x04;
            if(input_mask & KEY_PAD1_RIGHT)   temp &= ~0x08;
            if(input_mask & KEY_PAD1_BUTTON2) temp &= ~0x10;
            if(input_mask & KEY_PAD1_BUTTON1) temp &= ~0x20;
            if(input_mask & KEY_PAD2_UP)      temp &= ~0x40;
            if(input_mask & KEY_PAD2_DOWN)    temp &= ~0x80;
            return (temp);

        case 0xC1: /* INPUT #1 */
        case 0xDD:
            temp = 0xFF;
            if(input_mask & KEY_PAD2_LEFT)    temp &= ~0x01;
            if(input_mask & KEY_PAD2_RIGHT)   temp &= ~0x02;
            if(input_mask & KEY_PAD2_BUTTON2) temp &= ~0x04;
            if(input_mask & KEY_PAD2_BUTTON1) temp &= ~0x08;
            if(input_mask & KEY_PAD_RESET)    temp &= ~0x10;
            return (temp);

        case 0xBE: /* VDP DATA */
            return (vdp_data_r());

        case 0xBD:
        case 0xBF: /* VDP CTRL */
            return (vdp_ctrl_r());

        case 0xF2: /* YM2413 DETECT */
            break;

        default:
            break;
    }

    return (0xFF);
}


void change_pc16(int pc)
{
}


void change_pc(int pc)
{
}


int cpu_getactivecpu(void)
{
    return (0);
}


int irq_callback(int param)
{
    return (0xFF);
}


void sms_mapper_w(int address, int data)
{
    /* Calculate ROM page index */
    byte page = (data % cart.pages);

    /* Save frame control register data */
    sms.fcr[address] = data;

    switch(address)
    {
        case 0:
            if(data & 8)
            {
                sms.save = 1;
                sms.read_map[4]  = &sms.sram[(data & 4) ? 0x4000 : 0x0000];
                sms.read_map[5]  = &sms.sram[(data & 4) ? 0x6000 : 0x2000];
                sms.write_map[4] = &sms.sram[(data & 4) ? 0x4000 : 0x0000];
                sms.write_map[5] = &sms.sram[(data & 4) ? 0x6000 : 0x2000];
            }
            else
            {
                sms.read_map[4]  = &cart.rom[((sms.fcr[3] % cart.pages) * 0x4000) + 0x0000];
                sms.read_map[5]  = &cart.rom[((sms.fcr[3] % cart.pages) * 0x4000) + 0x2000];
                sms.write_map[4] = sms.dummy;
                sms.write_map[5] = sms.dummy;
            }
            break;

        case 1:
            sms.read_map[0] = &cart.rom[(page * 0x4000) + 0x0000];
            sms.read_map[1] = &cart.rom[(page * 0x4000) + 0x2000];
            break;

        case 2:
            sms.read_map[2] = &cart.rom[(page * 0x4000) + 0x0000];
            sms.read_map[3] = &cart.rom[(page * 0x4000) + 0x2000];
            break;

        case 3:
            if(!(sms.fcr[0] & 0x08))
            {
                sms.read_map[4] = &cart.rom[(page * 0x4000) + 0x0000];
                sms.read_map[5] = &cart.rom[(page * 0x4000) + 0x2000];
            }
            break;
    }
}

